#include <iostream>
#include <blt/std/allocator.h>
#include <blt/gfx/window.h>
#include <blt/gfx/state.h>
#include <blt/gfx/stb/stb_image.h>
#include <blt/std/hashmap.h>
#include <functions.h>
#include "blt/gfx/renderer/resource_manager.h"
#include "blt/gfx/renderer/batch_2d_renderer.h"
#include "blt/std/assert.h"
#include "imgui.h"
#include <variant>
#include <random>

blt::gfx::matrix_state_manager global_matrices;
blt::gfx::resource_manager resources;
blt::gfx::batch_renderer_2d renderer_2d(resources);

struct node;

blt::area_allocator<node, 32000> node_allocator;

node* createNode(function_t type)
{
    auto* n = node_allocator.allocate(1);
    node_allocator.construct(n, type);
    return n;
}


void destroyNode(node* n)
{
    node_allocator.destroy(n);
    node_allocator.deallocate(n, 1);
}

class tree;

struct node
{
        friend tree;
    private:
        // only valid up to the argc
        std::array<node*, MAX_ARGS> sub_nodes;
        size_t argc = 0;
        
        function_t type = function_t::ADD;
        data_t data;
        
        std::optional<image> img;
        
        void grow()
        {
            auto min_children = function_arg_min_map[to_underlying(type)];
            auto max_children = function_arg_max_map[to_underlying(type)];
            
            static thread_local std::random_device dev;
            static thread_local std::mt19937_64 engine{dev()};
            std::uniform_int_distribution<int> dist(min_children, max_children);
            
            argc = dist(engine);
            populate_sub_nodes();
        }
        
        void full()
        {
            argc = function_arg_max_map[to_underlying(type)];
            populate_sub_nodes();
        }
        
        void populate_sub_nodes()
        {
            const auto& allowed_args = function_arg_allowed_map[to_underlying(type)];
            static thread_local std::random_device dev;
            static thread_local std::mt19937_64 engine{dev()};
            std::uniform_int_distribution<int> dist(0, allowed_args.size() - 1);
            for (size_t i = 0; i < argc; i++)
            {
                // each arg can have a unique set of allowed args
                sub_nodes[i] = createNode(allowed_args[i][dist(engine)]);
            }
        }
    
    public:
        node(function_t type): type(type)
        {
            static thread_local std::random_device dev;
            static thread_local std::mt19937_64 engine{dev()};
            static thread_local std::uniform_real_distribution<float> dist(0.0f, 1.0f);
            if (type == function_t::SCALAR)
                data[0] = dist(engine);
            else if (type == function_t::COLOR)
            {
                for (auto& v : data)
                    v = dist(engine);
            }
        }
        
        static node* construct_random_tree(const size_t depth = 0)
        {
            node* n = createNode(function_t::NOISE);
            
            std::random_device dev;
            std::mt19937_64 engine{dev()};
            std::uniform_real_distribution<float> dist(0.0f, 1.0f);
            for (float& f : n->data)
                f = dist(engine);
            
            return n;
        }
        
        void evaluate()
        {
            img = image{};
            for (blt::i32 y = 0; y < height; y++)
            {
                for (blt::i32 x = 0; x < width; x++)
                {
                
                }
            }
        }
        
        void evaluate_tree()
        {
            for (size_t i = 0; i < argc; i++)
                sub_nodes[i]->evaluate_tree();
            evaluate();
        }
        
        image& getImage()
        {
            return img.value();
        }
        
        ~node()
        {
            for (auto* node : sub_nodes)
                destroyNode(node);
        }
};

class tree
{
    public:
        node* root;
    
    
};

node* root;
blt::gfx::texture_gl2D* texture;

bool fucky(float f)
{
    return f == std::numeric_limits<float>::quiet_NaN() || f == -std::numeric_limits<float>::quiet_NaN() ||
           f == std::numeric_limits<float>::infinity() ||
           f == -std::numeric_limits<float>::infinity();
}

void init()
{
    global_matrices.create_internals();
    renderer_2d.create();
    
    
    texture = new blt::gfx::texture_gl2D(width, height);
    resources.set("img", texture);
    resources.enqueue("../libraries/BLT-With-Graphics-Template/resources/textures/parkerfemBOY.png", "cum");
    
    resources.load_resources();
}

void update(std::int32_t w, std::int32_t h)
{
    global_matrices.update_perspectives(w, h, 90, 0.1, 2000);
    glDisable(GL_BLEND);
    
    if (ImGui::Button("Regenerate"))
    {
        BLT_INFO("Regen tree");
        destroyNode(root);
        root = node::construct_random_tree();
        root->evaluate_tree();
        BLT_INFO("Preprocess");
        auto& data = root->getImage()->getData();
        for (auto& v : data)
        {
            //BLT_INFO_STREAM << "B4: " << v.x() << " " << v.y() << " " << v.z() << " " << v.w() << "\n";
            if (fucky(v.x()))
                v[0] = 0;
            if (fucky(v.y()))
                v[1] = 0;
            if (fucky(v.z()))
                v[2] = 0;
            v = v.normalize();
            if (fucky(v.x()))
                v[0] = 0;
            if (fucky(v.y()))
                v[1] = 0;
            if (fucky(v.z()))
                v[2] = 0;
            //BLT_INFO_STREAM << "A4: " << v.x() << " " << v.y() << " " << v.z() << " " << v.w() << "\n";
        }
        
        BLT_INFO("Uploading");
        texture->upload((void*) root->getImage()->getData().data(), GL_RGB, 0, 0, 0, -1, -1, GL_FLOAT);
        resources.set("img", texture);
    }
    
    auto lw = 512.0f;
    auto lh = 512.0f;
    //renderer_2d.drawRectangle(blt::vec4{0.5, 0.0, 1.0, 1.0},
    //                          {static_cast<float>(w) / 2.0f, static_cast<float>(h) / 2.0f, static_cast<float>(w), static_cast<float>(h)});
    renderer_2d.drawRectangle("cum", {static_cast<float>(w) / 2.0f, static_cast<float>(h) / 2.0f, lw, lh});
    
    global_matrices.update();
    renderer_2d.render();
}

int main()
{
    blt::gfx::init(blt::gfx::window_data{"Window of GP test", init, update}.setSyncInterval(1));
    global_matrices.cleanup();
    resources.cleanup();
    renderer_2d.cleanup();
    blt::gfx::cleanup();
    return 0;
}
